MODULE builder

  ! tool to build up executable programs or libraries using a configuration file
  
  USE odessa
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: builddir,nerror,errmes,analbase,echo
  PUBLIC :: build_init,build_task,print_manual
  
  CHARACTER      :: cdir         ! directory delimiter
  CHARACTER(255) :: builddir     ! name of the build directory
  TYPE(odbase)   :: platform     ! selected platform sub-base of config
  TYPE(odbase)   :: rootcfg      ! the configuration base
  CHARACTER(8)   :: obj_suffix   ! OS dependent suffix for object files
  CHARACTER(8)   :: mod_suffix   ! OS dependent suffix for module files
  CHARACTER(8)   :: lib_suffix   ! OS dependent suffix for library files
  CHARACTER(8)   :: dyn_suffix   ! OS dependent suffix for dynamic library files
  CHARACTER(8)   :: lib_prefix   ! OS dependent prefix for library files
  CHARACTER(8)   :: dyn_prefix   ! OS dependent prefix for dynamic library files
  CHARACTER(8)   :: exe_suffix   ! OS dependent suffix for program files
  TYPE(odbase)   :: analbase     ! the base containing analyzer variables,
                                 ! in particular "config"
  INTEGER        :: echo=0       ! echo ouput unit (useful in debug mode)
                                 ! <= 0 => no echo
  INTEGER        :: nerror=0
  CHARACTER(150),ALLOCATABLE :: errmes(:) ! error messages
  
  CONTAINS
  
  SUBROUTINE build_init(config,cplatform)
  
    TYPE(odbase),INTENT(inout) :: config
    CHARACTER(*),INTENT(in)    :: cplatform
    
    INTEGER        :: k,ideb,ifin
    CHARACTER(255) :: cplat,cp
    TYPE(odbase)   :: except
    
    ALLOCATE(errmes(256))
    
    rootcfg=config

    IF(cplatform == ' ') THEN
      CALL odget(config,'PLATFORM',platform)
    ELSE
      CALL odget(config,'PLATFORM',cplatform,platform)
    ENDIF
    IF(.NOT.odisvalid(platform)) THEN
      STOP "builder : invalid platform set"
    ENDIF
    CALL odget(platform,'NAME',cplat)
    cdir=odcdir()
    
    DO k=odcard(config,'EXCEPT'),1,-1
      CALL odget(config,'EXCEPT',except,k)
      CALL odget(except,'PLATFORM',cp)
      IF(cp /= cplat) THEN
        CALL oddelete(config, 'EXCEPT',k)
      ENDIF
    ENDDO

    CALL odget(platform,'OBJ',obj_suffix)
    CALL odget(platform,'MOD',mod_suffix)
    CALL odget(platform,'LIB',lib_suffix)
    CALL odget(platform,'DYN',dyn_suffix)
    CALL odget(platform,'EXE',exe_suffix)
    CALL odget(platform,'BUILD',builddir)
    CALL odget(platform,'LIBPRE',lib_prefix)
    CALL odget(platform,'DYNPRE',dyn_prefix)
    
    IF(.NOT. odfile_exist(builddir)) THEN
    
      ideb=1
      ifin=0
      DO
        k=index(builddir(ideb:),cdir)
        IF(k <= 0) EXIT
        IF(k > 0) THEN
          ifin=ideb+k-2
          IF(.NOT. odfile_exist(builddir(1:ifin))) THEN
            k=odfile_mkdir(builddir(1:ifin))
            IF(k /= 0) THEN
              CALL error('impossible to create',builddir(1:ifin))
              STOP 'stop of the builder : too strong error'
            ENDIF
          ENDIF
          ideb=k+2
        ENDIF
      ENDDO
      
      k=odfile_mkdir(builddir)
      IF(k /= 0) THEN
        CALL error('impossible to create',builddir(1:ifin))
        STOP 'stop of the builder : too strong error'
      ENDIF
      
    ENDIF
  
    ! initializing the analyzer data base
    
    CALL odinit(analbase)
    CALL odinsert(analbase,'config',config)
    CALL odinsert(analbase,'platform',platform)
    
    ! common suffixes if missing

    IF(obj_suffix == ' ') THEN
      IF(cdir == '/') THEN
        obj_suffix='.o'
      ELSE
        obj_suffix='.obj'
      ENDIF
    ENDIF
    
    IF(mod_suffix == ' ') mod_suffix='.mod'
    IF(exe_suffix == ' ') exe_suffix='.exe'
    
    IF(lib_suffix == ' ') THEN
      IF(cdir == '/') THEN
        lib_suffix='.a'
      ELSE
        lib_suffix='.lib'
      ENDIF
    ENDIF
    
  END SUBROUTINE build_init
  
  RECURSIVE SUBROUTINE build_task(config,build,status,niv)
  
    TYPE(odbase),INTENT(inout) :: config ! configuration of the current tasks
    TYPE(odbase),INTENT(inout) :: build  ! current build
    INTEGER     ,INTENT(out)   :: status ! output status
    INTEGER     ,INTENT(in)    :: niv    ! task level
    
    INTEGER :: ntask,k,l,n,istat
    INTEGER,ALLOCATABLE :: level(:),stat(:),list(:)
    TYPE(odbase),POINTER :: task(:) ! array of sub-tasks
    TYPE(odbase)         :: btask   ! same task during a previous build
    CHARACTER(32)  :: cname
    CHARACTER(255) :: bdir,cdir
    
    CALL execute_conditions(config) ! solving COND blocks
    
    status=0
    
    CALL odgetptrfamily(config,'TASK',task)
    ntask=SIZE(task)

    ALLOCATE(list(ntask),level(ntask),stat(ntask))
    
    CALL compute_task_level(config,build,level)
  
    DO l=1,MAXVAL(level)
      n=0
      DO k=1,ntask
        IF(level(k) == l) THEN
          n=n+1
          list(n)=k
        ENDIF
      ENDDO
      
      DO k=1,n
        CALL odget(task(list(k)),'NAME',cname)
        CALL odget(build,'TASK',cname,btask)
        IF(.NOT.odisvalid(btask)) THEN
          CALL odinit(btask)
          CALL odput(btask,'NAME',cname)
          CALL odinsert(build,'TASK',btask)
        ENDIF
        CALL build_task(task(list(k)),btask,stat(k),niv+1)
      ENDDO
      
      status=SUM(stat(1:n))
      
      IF(status /= 0) THEN
        ! the tasks which have not been treated must get their old status
        DO k=1,ntask
          IF(level(k) > l) THEN
            CALL odget(task(k),'NAME',cname)
            CALL odget(build,'TASK',cname,btask)
            IF(odisvalid(btask)) THEN
              CALL odput(config,'TASK',btask,k) ! replacing the task by the previous one
            ENDIF
          ENDIF
        ENDDO
        GOTO 100
      ENDIF

    ENDDO

    IF(niv > 1 .AND. odcard(config,'BUILD') > 0) THEN
      CALL odget(config,'BUILD',cdir)
      CALL odfile_name(bdir,builddir,cdir)    
      IF(.NOT.odfile_exist(bdir)) THEN
        istat=odfile_mkdir(bdir)
        IF(istat /= 0) THEN
          CALL error('impossible to build the directory',bdir)
          STOP "too strong error"
        ENDIF
      ENDIF
    ENDIF
    
    CALL build_obj(platform,config,build,status,niv) ! compiling internal source files
    IF(status /= 0) GOTO 100
      
    CALL build_lib(config,status,niv)                ! making the optional library
    IF(status /= 0) GOTO 100
      
    CALL build_dyn(config,status,niv)                ! making the optional dynamic library
    IF(status /= 0) GOTO 100
    
    CALL build_exe(config,status,niv)                ! making the optional executable program
    IF(status /= 0) GOTO 100
    
    CALL build_inst(config,status,niv)               ! executing additional instructions
    IF(status /= 0) GOTO 100
    
    CALL odget(config,'NAME',cname)
    IF(cname /= ' ') &
    CALL message(niv-1,'up-to-date task',cname)
    RETURN
    
    100 CONTINUE
    
    CALL odget(config,'NAME',cname)
    IF(cname /= ' ') THEN
      CALL message(niv-1,'task failure',cname)
      CALL error('task failure',cname)
    ENDIF
    RETURN
    
  END SUBROUTINE build_task
  
  SUBROUTINE execute_conditions(config)
    TYPE(odbase),INTENT(inout) :: config ! current configuration
    INTEGER :: ir
    TYPE(odelem) :: result
    TYPE(odbase) :: cond
    DO
      CALL odget(config,'COND',cond)
      IF(.NOT.odisvalid(cond)) EXIT
      CALL odanalyzer(analbase,cond,echo,result)
      IF(odtype(result) == od_i0) THEN
        ir=result
        IF(ir == 1) THEN
          CALL oddelete(cond,'IF')
          CALL odappend(config,cond)
          CALL oddelete(cond)
        ENDIF
      ENDIF
    ENDDO
  END SUBROUTINE execute_conditions
  
  SUBROUTINE compute_task_level(config,build,level)
  
    TYPE(odbase),INTENT(in)    :: config
    TYPE(odbase),INTENT(inout) :: build
    INTEGER     ,INTENT(out)   :: level(:)
    INTEGER :: i,j,k,lold,ntask,lnew
    TYPE(odbase),POINTER :: task(:)
    CHARACTER(32) :: taskname(SIZE(level)),cdep
    INTEGER,POINTER :: list(:)
    TYPE(odbase) :: btask
    
    level=1
    CALL odgetptrfamily(config,'TASK',task)
    
    ntask=SIZE(task)
    
    DO i=ntask,1,-1
      CALL odget(task(i),'NAME',taskname(i))
      ! synchronizing tasks in config and build
      CALL odget(build,'TASK',taskname(i),btask)
      IF(.NOT.odisvalid(btask)) THEN
        CALL odinit(btask)
        CALL odinsert(btask,'NAME',taskname(i))
        CALL odinsert(build,'TASK',btask)
      ENDIF
    ENDDO
    
    ! ordering tasks
    
    DO i=1,ntask
    
      loopj : DO j=1,odcard(task(i),'DEPEND')
        CALL odget(task(i),'DEPEND',cdep,j)
        DO k=1,ntask
          IF(cdep == taskname(k)) THEN
            CALL odinsert(task(i),'TASKDEP',k)
            CYCLE loopj
          ENDIF
        ENDDO
        CALL error('task',cdep,'not found')
      ENDDO loopj
      
    ENDDO
    
    lold=MAXVAL(level)
    
    DO
      DO i=1,ntask
        CALL odgetptrfamily(task(i),'TASKDEP',list)
        DO j=1,SIZE(list)
          level(i)=MAX(level(i),level(list(j))+1)
        ENDDO
      ENDDO
      lnew=MAXVAL(level)
      IF(lnew == lold) RETURN
      IF(lnew > ntask) THEN
        WRITE(*,*) 'BUILDER : recursivity detected between tasks'
        STOP
      ENDIF
      lold=lnew
    ENDDO
    
  END SUBROUTINE compute_task_level
  
  SUBROUTINE build_obj(config,task,build,status,niv)
  
    ! compiling source files
  
    TYPE(odbase),INTENT(inout) :: config ! current set of tasks
    TYPE(odbase),INTENT(inout) :: task   ! current task
    TYPE(odbase),INTENT(inout) :: build  ! previous build of that task
    INTEGER     ,INTENT(out)   :: status ! final status : 0=OK
    INTEGER     ,INTENT(in)    :: niv

    INTEGER                  :: i,k,l,depstat,nfile,ns,icomp,n,ncomp
    CHARACTER(512)           :: cdir,fullname
    CHARACTER(255)           :: dirname,basename,suffix
    INTEGER,ALLOCATABLE      :: level(:),list(:),liststatus(:),listcomp(:)
    INTEGER,POINTER          :: listsubtask(:)
    TYPE(odbase)             :: file,dir
    TYPE(odbase),POINTER     :: listfile(:)
    CHARACTER(8),ALLOCATABLE :: validsuffix(:)
    
    status = 0
    
    ! constructing the list of source files of the task
    
    CALL odgetcopyfamily(task,'SUFFIX',validsuffix)
    
    DO k=1,odcard(task,'DIR')
      CALL odget(task,'DIR',cdir,k)
      CALL odinit(dir)
      CALL odfile_listdir(dir,cdir)
      DO i=1,odcard(dir,'FILE')
        CALL odget(dir,'FILE',fullname,i)
        CALL odfile_splitname(fullname,dirname,basename,suffix)
        DO l=1,SIZE(validsuffix)
          IF(suffix == validsuffix(l)) THEN
            CALL odfile_name(fullname,cdir,TRIM(basename)//TRIM(suffix))
            CALL odinit(file)
            CALL odput(file,'NAME',fullname)
            CALL odinsert(task,'FILE',file)
            EXIT
          ENDIF
        ENDDO
      ENDDO
      CALL oddelete(dir)
    ENDDO
    
    ! explicit list of file names
    
    DO k=1,odcard(task,'FILENAME')
      CALL odget(task,'FILENAME',fullname,k)
      CALL odget(task,'FILE',fullname,file)
      IF(.NOT.odisvalid(file)) THEN
        CALL odinit(file)
        CALL odput(file,'NAME',fullname)
        CALL odinsert(task,'FILE',file)
      ENDIF
    ENDDO
    
    ! Getting the final list of source files
    
    CALL odgetptrfamily(task,'FILE',listfile)
    nfile=SIZE(listfile)
    
    ! constructing the list of dependencies and the list of generated files
    
    DO i=1,nfile
      CALL list_target_depend(task,build,listfile(i),status,niv)
      if(status /= 0) RETURN
    ENDDO
    
    ! List of files associated to a job and list of all targets
    
    DO i=1,nfile
      IF(odcard(listfile(i),'TARGET') > 0) THEN
        CALL odinsert(task,'SUBTASK',i)
      ENDIF
    ENDDO

    ! checking the external dependencies : all external dependencies must be present somewhere
    
    CALL check_external_dependencies(config,task,status)
    
    IF(status /= 0) RETURN
    
    ! computing the level of files, files having the lowest level being compiled first
    
    ALLOCATE(level(nfile))
    CALL compute_file_level(task,level,status)
    
    IF(status /= 0) RETURN
    
    ! computing the generation of target files
    
    CALL odgetptrfamily(task,'SUBTASK',listsubtask)
    
    n=SIZE(listsubtask)
    
    ALLOCATE(list(n),liststatus(n),listcomp(n))
    
    DO k=1,MAXVAL(level)
    
      ! list of sub-tasks having the level k
      
      ns=0
      
      DO i=1,SIZE(listsubtask)
        IF(level(listsubtask(i)) == k) THEN
          ns=ns+1
          list(ns)=listsubtask(i)
        ENDIF
      ENDDO
      
      ! all the subtasks with the same level may be compiled in //
      
      liststatus(1:ns)=0
      
      ncomp=0
      
      DO i=1,ns
        file=listfile(list(i))
        CALL odget(file,'COMPILED',icomp)
        IF(icomp > 0) THEN ! already compiles during a previous build
          CALL check_dependency(task,file,depstat,niv)
          IF(depstat == 0) CYCLE
        ELSE
          IF(echo > 0) THEN
            CALL odget(file,'NAME',fullname)
            IF(icomp == 0) THEN
              CALL message(niv,'compiling',fullname,'because never compiled up to now')
            ELSE
              CALL message(niv,'recompiling',fullname,'because of modification')
            ENDIF
          ENDIF
        ENDIF
        ncomp=ncomp+1
        listcomp(ncomp)=list(i)
      ENDDO
      !$OMP PARALLEL DO SCHEDULE(static,2)
      DO i=1,ncomp
        CALL compile_file(task,listfile(listcomp(i)),liststatus(i),niv)
      ENDDO
      !$OMP END PARALLEL DO
      status=status+SUM(liststatus(1:ncomp))
      IF(status /= 0) RETURN
      
    ENDDO
    
  END SUBROUTINE build_obj
  
  SUBROUTINE build_lib(task,status,niv)
  
    TYPE(odbase),INTENT(inout) :: task
    INTEGER     ,INTENT(out)   :: status
    INTEGER     ,INTENT(in)    :: niv
    
    CHARACTER(255)             :: name,flags
    CHARACTER(512)             :: fullname,libname,dirname,filename
    CHARACTER(32)              :: out
    CHARACTER(8)               :: ctype
    INTEGER                    :: k,n,i,j,time,oldtime,icomp,mode,n0
    TYPE(odbase)               :: file,target,libopt,stask,dirbase
    CHARACTER(255)             :: command
    CHARACTER(50000)           :: arguments
    CHARACTER                  :: c
    
    status=0
    
    IF(odcard(task,'LIBNAME') == 0) RETURN

    CALL odget(task,'LIBNAME',name)
    CALL odfile_name(libname,builddir,TRIM(lib_prefix)//TRIM(name)//lib_suffix)
    CALL odfile_info(libname,time=oldtime) ! time stamp of the library
   
    ! building the command to update the library
    
    CALL odget(platform,'LIBRARY',libopt)
    CALL odget(libopt,'COMMAND',command)
    CALL odget(libopt,'OUT'    ,out)
    CALL odget(libopt,'MODE'   ,mode)
    CALL odget(libopt,'FLAGS'  ,flags)
    n=LEN_TRIM(out) ; c=out(n:n)
    IF(c == '_') THEN
      n=n-1
    ELSE IF(c /= ':' .AND. c/= '=') THEN
      n=n+1
    ENDIF
    
    IF(mode == 2) THEN
      IF(oldtime == 0) THEN
        arguments=TRIM(flags)//' '//out(1:n)//TRIM(libname)
      ELSE
        arguments=TRIM(flags)//' '//out(1:n)//TRIM(libname)//' '//TRIM(libname)
      ENDIF
    ELSE
      arguments=TRIM(flags)//' '//out(1:n)//TRIM(libname)
    ENDIF
    
    ! getting names of object files which are more recent than the library
    
    n=0
   
    DO j=0,odcard(task,'TASK')
      IF(j == 0) THEN
        stask=task
      ELSE
        CALL odget(task,'TASK',stask,j)
      ENDIF
      DO i=1,odcard(stask,'SUBTASK')
        CALL odget(stask,'SUBTASK',k,i)
        CALL odget(stask,'FILE',file,k)
        CALL odget(file,'TARGET',target)
        CALL odget(target,'TYPE',ctype)
        IF(ctype /= 'OBJECT') CYCLE
        CALL odget(file,'COMPILED',icomp)
        IF(icomp > 0) THEN
          CALL odget(target,'FULLNAME',fullname)
          IF(fullname(LEN_TRIM(fullname)-LEN_TRIM(obj_suffix)+1:) /= obj_suffix) CYCLE
          CALL odfile_info(fullname,time=time)
          IF(time > oldtime) THEN
            n=n+1
            arguments=TRIM(arguments)//' '//TRIM(fullname)
          ENDIF
        ENDIF
      ENDDO
    ENDDO
    
    DO j=1,odcard(task,'DIROBJ')
      n0=n
      CALL odget(task,'DIROBJ',filename,j)
      CALL odfile_name(dirname,builddir,filename)
      CALL odinit(dirbase)
      CALL odfile_listdir(dirbase,dirname)
      DO k=1,odcard(dirbase,'FILE')
        CALL odget(dirbase,'FILE',filename,k)
        CALL odfile_name(fullname,dirname,filename)
        IF(fullname(LEN_TRIM(fullname)-LEN_TRIM(obj_suffix)+1:) /= obj_suffix) CYCLE
        CALL odfile_info(fullname,time=time)
        IF(time > oldtime) THEN
          n=n+1
          EXIT
        ENDIF
      ENDDO
      IF(n /= n0) THEN
        arguments=TRIM(arguments)//' '//TRIM(dirname)//odcdir()//'*'//obj_suffix
      ENDIF
      CALL oddelete(dirbase)
    ENDDO
    
    IF(n == 0) RETURN
    
    ! Executing the command
    
    CALL message (niv,command,arguments)
    CALL odsystem(command,arguments,status)
    CALL odfile_info(libname,time=time)
    IF(time <= oldtime.OR.status /= 0) THEN
      CALL error('impossible to create the library',libname)
      status=1
      RETURN
    ENDIF
    
  END SUBROUTINE build_lib
  
  SUBROUTINE build_dyn(task,status,niv)
  
    TYPE(odbase),INTENT(inout) :: task
    INTEGER     ,INTENT(out)   :: status
    INTEGER     ,INTENT(in)    :: niv
    
    CHARACTER(255)             :: libname,flags,progname,tail
    CHARACTER(512)             :: fullname,dynname,dirname,filename
    CHARACTER(32)              :: out
    CHARACTER(8)               :: ctype
    INTEGER                    :: k,n,i,j,time,oldtime,idyn,ifound
    TYPE(odbase)               :: file,target,dynopt,stask,dirbase
    CHARACTER                  :: c
    CHARACTER(255)             :: command
    CHARACTER(10000)           :: arguments
    
    status=0
    
    IF(odcard(task,'DYNNAME') == 0) RETURN
    
    CALL odget(task,'DYNNAME',fullname)
    CALL odfile_name(dynname,builddir,TRIM(dyn_prefix)//TRIM(fullname)//dyn_suffix)
    CALL odfile_info(dynname,time=oldtime) ! time stamp of the executable program
    
    idyn=0
    IF(oldtime == 0) idyn=1
    
    ! building the command head to update the executable program
    
    CALL odget(platform,'DYNLIB',dynopt)
    CALL odget(dynopt,'COMMAND',command)
    CALL odget(dynopt,'OUT'    ,out)
    CALL odget(dynopt,'FLAGS'  ,flags)
    n=LEN_TRIM(out) ; c=out(n:n)
    IF(c == '_') THEN
      n=n-1
    ELSE IF(c /= ':' .AND. c /= '=') THEN
      n=n+1
    ENDIF
    
    arguments=TRIM(flags)//' '//out(1:n)//TRIM(dynname)
    
    ! getting the object files
   
    DO j=0,odcard(task,'TASK')
      IF(j == 0) THEN
        stask=task
      ELSE
        CALL odget(task,'TASK',stask,j)
      ENDIF
      DO k=1,odcard(stask,'SUBTASK')
        CALL odget(stask,'SUBTASK',i,k)
        CALL odget(stask,'FILE',file,i)
        CALL odget(file,'TARGET',target)
        CALL odget(target,'TYPE',ctype)
        IF(ctype /= 'OBJECT') CYCLE
        CALL odget(target,'FULLNAME',fullname)
        IF(fullname(LEN_TRIM(fullname)-LEN_TRIM(obj_suffix)+1:) /= obj_suffix) CYCLE
        CALL odget(target,'TIME',time)
        IF(time > oldtime) idyn=idyn+1
        arguments=TRIM(arguments)//' '//TRIM(fullname)
      ENDDO
    ENDDO
    
    DO j=1,odcard(task,'DIROBJ')
      CALL odget(task,'DIROBJ',filename,j)
      CALL odfile_name(dirname,builddir,filename)
      CALL odinit(dirbase)
      CALL odfile_listdir(dirbase,dirname)
      n=0
      DO k=1,odcard(dirbase,'FILE')
        CALL odget(dirbase,'FILE',filename,k)
        CALL odfile_name(fullname,dirname,filename)
        IF(fullname(LEN_TRIM(fullname)-LEN_TRIM(obj_suffix)+1:) /= obj_suffix) CYCLE
        n=n+1
        CALL odfile_info(fullname,time=time)
        IF(time > oldtime) THEN
          idyn=idyn+1
          EXIT
        ENDIF
      ENDDO
      IF(n > 0) THEN
        arguments=TRIM(arguments)//' '//TRIM(dirname)//odcdir()//'*'//obj_suffix
      ENDIF
      CALL oddelete(dirbase)
    ENDDO
    
    ! test of modification
    
    IF(idyn == 0) RETURN ! not modified

    ! Executing the command and updating the time stamp
    
    CALL message (niv,command,arguments)
    CALL odsystem(command,arguments,status)
    CALL odfile_info(dynname,time=time)
    IF(time <= oldtime.OR.status /= 0) THEN
      CALL error('impossible to create the dynamic library',dynname)
      status=1
      RETURN
    ENDIF
    
  END SUBROUTINE
  
  SUBROUTINE build_exe(task,status,niv)
  
    TYPE(odbase),INTENT(inout) :: task
    INTEGER     ,INTENT(out)   :: status
    INTEGER     ,INTENT(in)    :: niv
    
    CHARACTER(255)             :: libname,flags,progname,tail
    CHARACTER(512)             :: fullname,exename
    CHARACTER(32)              :: out
    CHARACTER(8)               :: ctype
    INTEGER                    :: k,n,i,time,oldtime,iexe,ifound
    TYPE(odbase)               :: file,target,exeopt
    CHARACTER                  :: c
    CHARACTER(255)             :: command
    CHARACTER(10000)           :: arguments
    
    status=0
    
    IF(odcard(task,'PROGRAM') == 0) RETURN
    
    CALL odget(task,'PROGRAM',progname)
    CALL odfile_name(exename,builddir,TRIM(progname)//exe_suffix)
    CALL odfile_info(exename,time=oldtime) ! time stamp of the executable program
    
    iexe=0
    IF(oldtime == 0) iexe=1
    
    ! building the command head to update the executable program
    
    CALL odget(platform,'PROGRAM',exeopt)
    CALL odget(exeopt,'COMMAND',command)
    CALL odget(exeopt,'OUT'    ,out)
    CALL odget(exeopt,'FLAGS'  ,flags)
    n=LEN_TRIM(out) ; c=out(n:n)
    IF(c == '_') THEN
      n=n-1
    ELSE IF(c /= ':' .AND. c /= '=') THEN
      n=n+1
    ENDIF
    
    arguments=TRIM(flags)//' '//out(1:n)//TRIM(exename)
    
    ! getting the object files

    DO k=1,odcard(task,'SUBTASK')
      CALL odget(task,'SUBTASK',i,k)
      CALL odget(task,'FILE',file,i)
      CALL odget(file,'TARGET',target)
      CALL odget(target,'TYPE',ctype)
      IF(ctype /= 'OBJECT') CYCLE
      CALL odget(target,'FULLNAME',fullname)
      IF(fullname(LEN_TRIM(fullname)-LEN_TRIM(obj_suffix)+1:) /= obj_suffix) CYCLE
      CALL odget(target,'TIME',time)
      IF(time > oldtime) iexe=iexe+1
      arguments=TRIM(arguments)//' '//TRIM(fullname)
    ENDDO
    
    ! getting the libraries

    DO k=1,odcard(task,'LIBRARY')
      CALL odget(task,'LIBRARY',libname,k)
      CALL build_search(task,TRIM(lib_prefix)//TRIM(libname)//lib_suffix,fullname,ifound)
      IF(ifound == 0) THEN
        CALL local_search(task,TRIM(lib_prefix)//TRIM(libname)//lib_suffix,fullname,ifound)
        IF(ifound == 0) THEN
          CALL external_search(TRIM(lib_prefix)//TRIM(libname)//lib_suffix,fullname,ifound)
          IF(ifound == 0) THEN
            CALL error("impossible to find the library",TRIM(lib_prefix)//TRIM(libname)//lib_suffix)
            status=1
            RETURN
          ENDIF
        ENDIF
      ENDIF
      CALL odfile_info(fullname,time=time)
      IF(time > oldtime) iexe=iexe+1
      arguments=TRIM(arguments)//' '//TRIM(fullname)
    ENDDO
    
    ! getting the dynamic libraries

    DO k=1,odcard(task,'DYNLIB')
      CALL odget(task,'DYNLIB',libname,k)
      CALL build_search(task,TRIM(dyn_prefix)//TRIM(libname)//dyn_suffix,fullname,ifound)
      IF(ifound == 0) THEN
        CALL local_search(task,TRIM(dyn_prefix)//TRIM(libname)//dyn_suffix,fullname,ifound)
        IF(ifound == 0) THEN
          CALL external_search(TRIM(dyn_prefix)//TRIM(libname)//dyn_suffix,fullname,ifound)
          IF(ifound == 0) THEN
            CALL error("impossible to find the dynamic library ",TRIM(dyn_prefix)//TRIM(libname)//dyn_suffix)
            status=1
            RETURN
          ENDIF
        ENDIF
      ENDIF
      CALL odfile_info(fullname,time=time)
      IF(time > oldtime) iexe=iexe+1
      arguments=TRIM(arguments)//' '//TRIM(fullname)
    ENDDO
    
    ! test of modification
    
    IF(iexe == 0) RETURN ! not modified
    
    ! tail part of the command
    
    CALL odget(exeopt,'TAIL',tail)
    arguments=TRIM(arguments)//' '//tail
    
    ! Executing the command and updating the time stamp
    
    CALL message (niv,command,arguments)
    CALL odsystem(command,arguments,status)
    CALL odfile_info(exename,time=time)
    IF(time <= oldtime.OR.status /= 0) THEN
      CALL error('impossible to create the program',exename)
      status=1
      RETURN
    ENDIF
    
  END SUBROUTINE
  
  SUBROUTINE build_inst(task,status,niv)
  
    ! Executing a set of instructions 
    
    TYPE(odbase),INTENT(in)  :: task
    INTEGER     ,INTENT(out) :: status
    INTEGER     ,INTENT(in)  :: niv
    
    INTEGER        :: ninst
    CHARACTER(255) :: inst0,inst,text
    TYPE(odt)      :: ot
    TYPE(odbase)   :: proc,panal
    TYPE(odelem)   :: result
    INTEGER        :: k,i,ic,iq,ip,kdeb,kfin
    
    status=0
    ninst=odcard(task,'INST')
    IF(ninst == 0) RETURN
    
    CALL odinit(proc)
    CALL odinit(panal)
    
    CALL odput(panal,'platform',platform)
    CALL odput(panal,'task',task)
    
    DO i=1,ninst
      CALL odget(task,'INST',inst0,i)
      inst=inst0
      DO
        ic=0 ; iq=0 ; ip = 0 ; kdeb=0 ; kfin = 0
        DO k=1,LEN_TRIM(inst)
          IF(inst(k:k) == "'") THEN ! '
            IF(iq == 0) ic=1-ic
          ELSE IF(inst(k:k) == '"') THEN
            IF(ic == 0) iq=1-iq
          ELSE IF(inst(k:k) == "(") THEN
            IF(ic+iq == 0) THEN
              ip=ip+1
              IF(ip == 1) kdeb=k
            ENDIF
          ELSE IF(inst(k:k) == ")") THEN
            IF(ic+iq == 0) THEN
              ip=ip-1
              IF(ip == 0) THEN
                kfin=k
                EXIT
              ENDIF
            ENDIF
          ENDIF
        ENDDO
        IF(kdeb == 0) EXIT
        IF(kfin == 0) THEN
          CALL error('invalid analyzer instruction in :',inst0)
          inst(kdeb:)=' '
          EXIT
        ENDIF
        CALL odinit(ot,inst(kdeb:kfin))
        CALL odput(proc,'INST',ot)
        CALL odanalyser(panal,proc,echo=echo,result=result)
        IF(odtype(result) == od_c0 .OR. odtype(result) == od_t) THEN
          text=result
          inst=inst(1:kdeb-1)//TRIM(text)//inst(kfin+1:)
        ELSE
          inst=inst(1:kdeb-1)//inst(kfin+1:)
        ENDIF
      ENDDO
      IF(k == 1) CYCLE ! empty line
      CALL message(niv,inst)
      CALL odsystem(TRIM(inst))
    ENDDO
    
    CALL oddelete(panal)
    CALL oddelete(proc)
    
  END SUBROUTINE
  
  SUBROUTINE check_dependency(task,file,depstat,niv)
  
    TYPE(odbase),INTENT(in)    :: task    ! the task containing the file
    TYPE(odbase),INTENT(inout) :: file    ! current file to check
    INTEGER     ,INTENT(out)   :: depstat ! status of dependencies
    INTEGER     ,INTENT(in)    :: niv
    
    INTEGER        :: i,k,time,oldtime
    TYPE(odbase)   :: dep,depfile,target
    CHARACTER(512) :: fullname,filename
    INTEGER        :: sig,sigold
    CHARACTER(8)   :: ctype
    
    depstat=0
    
    IF(echo > 0) CALL odget(file,'NAME',filename) ! needed in verbose mode
    
    ! checking the targets first
    
    DO i=1,odcard(file,'TARGET')
      CALL odget(file,'TARGET',target,i)
      CALL odget(target,'FULLNAME',fullname)
      CALL odget(target,'TIME',oldtime)
      CALL odfile_info(fullname,time=time)
      IF(time < oldtime) THEN ! deleted target ?
        IF(echo > 0) THEN
          CALL odget(file,'NAME',filename)
          CALL message(niv,'recompiling',filename,'because the target',fullname,'is missing')
        ENDIF
        depstat=1
        RETURN
      ENDIF
    ENDDO
        
    DO i=1,odcard(file,'DEPEND')
      CALL odget(file,'DEPEND',dep,i)
      CALL odget(dep,'TIME',oldtime) ! dependency time stamp during another build
      IF(oldtime == 0) THEN
        IF(echo > 0) THEN
          CALL odget(dep,'FULLNAME',fullname)
          CALL message(niv,'recompiling',filename,'because the dependency',fullname,'has never been considered')
        ENDIF
        depstat=1
        RETURN
      ENDIF
      CALL odget(dep,'TYPE',ctype)
      CALL odget(dep,'DEPFILE',k)
      IF(ctype == 'INCLUDE' .AND. k > 0) THEN
        CALL odget(task,'FILE',depfile,k) ! the dependency is also a source file already compiled
        CALL odget(depfile,'TIME',time)
      ELSE
        CALL odget(dep,'FULLNAME',fullname)
        CALL odfile_info(fullname,time=time)
      ENDIF
      IF(time == oldtime) CYCLE ! the dependency has not changed
      IF(ctype == 'INCLUDE') THEN ! on do not check signature because an include can contain includes
        depstat=1                 ! => the signature is not correct in that case
        IF(echo > 0) THEN
          CALL message(niv,'recompiling',filename,'because the include file',fullname,'has a new date')
        ENDIF
        RETURN
      ENDIF
      CALL odget(dep,'SIG',sigold)
      IF(sigold == 0) THEN ! The signature does not exist
        depstat=1
        RETURN
      ENDIF
      CALL odfile_info(fullname,sig=sig)
      IF(sig /= sigold) THEN ! the dependency file has really changed
        IF(echo > 0) THEN
          CALL message(niv,'recompiling',filename,'because the used file',TRIM(fullname),'has a new signature')
        ENDIF
        depstat=1
        RETURN
      ENDIF
      CALL odput(dep,'TIME',time) ! just update the time stamp because the dependency file has not changed
    ENDDO

  END SUBROUTINE check_dependency
    
  SUBROUTINE list_target_depend(task,build,file,status,niv)
  
    TYPE(odbase),INTENT(in)    :: task  ! current task
    TYPE(odbase),INTENT(in)    :: build ! previous build
    TYPE(odbase),INTENT(inout) :: file  ! current file to analyze
    INTEGER     ,INTENT(out)   :: status ! =0 => OK
    INTEGER     ,INTENT(in)    :: niv
    
    CHARACTER(512) :: filename,dirname,line,fullname,bdir
    CHARACTER(255) :: comnew,comold
    CHARACTER(32)  :: basename
    CHARACTER      :: car
    CHARACTER(8)   :: suffix,ctype,cobj
    INTEGER        :: time,oldtime,k,istat,icar,ideb,ifin,kfile,imod,icomp
    INTEGER        :: sig,sigold
    TYPE(odbase)   :: oldfile,target,depend,compiler
    
    status=0
    
    CALL odget(file ,'NAME',filename)
    
    CALL odfile_splitname(filename,dirname,basename,suffix)
    CALL compute_command(filename,suffix,comnew)
    
    CALL odget(build,'FILE',filename,oldfile) ! getting the corresponding file in the previous build
    
    IF(odisvalid(oldfile)) THEN ! looking for a change in the source file compared to the previous build
      CALL odget(oldfile,'COMPILED',icomp)
      CALL odfile_info(filename,time=time)
      CALL odget(oldfile,'TIME',oldtime)
      IF(time /= oldtime) THEN
        CALL odget(oldfile,'SIG',sigold)
        CALL odfile_info(filename,sig=sig)
        IF(sig == sigold) THEN
          oldtime=time
          CALL odput(oldfile,'TIME',time) ! just updating the time stamp
        ENDIF
      ENDIF
      IF(time == oldtime) THEN ! no change : one replaces file by oldfile
        file=oldfile
        IF(icomp <= 0) THEN
          IF(echo > 0) &
          CALL message(niv,'the file',filename,'must be compiled because not compiled during the previous build')
        ELSE
          CALL odget(oldfile,'COMMAND',comold)
          IF(comold /= comnew) THEN
            CALL oddelete(file,'COMMAND')
            CALL odput(file,'COMMAND',comnew)
            IF(echo > 0) &
            CALL message(niv,'the file',filename,'must be compiled because the compilation command has changed')
            CALL odput(file,'COMPILED',-2)
          ELSE
            CALL odput(file,'COMPILED',2) ! to make a difference with current compilation (COMPILED=1)
          ENDIF
        ENDIF
        CALL odput(file,'COMMAND',comnew)
        RETURN
      ENDIF
    ELSE
      IF(echo > 0) CALL message(niv,'the file',filename,'has not been found in the previous build')
      CALL odput(file,'COMPILED',0) ! never compiled
    ENDIF

    CALL oddelete(file,'COMMAND')
    CALL odput(file,'COMMAND',comnew)
    CALL odput(file,'BASENAME',TRIM(basename))
    CALL odput(file,'DIRNAME',TRIM(dirname))
    CALL odput(file,'SUFFIX',TRIM(suffix))
    CALL odget(task,'BUILD',dirname)
    CALL odfile_name(bdir,builddir,dirname)
    
    ! One has now to analyze the contents of the source file, and therefore to open it first
    
    CALL odopen(unit=kfile,file=filename,status='OLD',FORM='FORMATTED')
    IF(kfile <= 0) THEN
      CALL error('impossible to open the file',filename)
      status=1
      RETURN
    ENDIF
    
    CALL odget(platform,'COMPILER','SUFFIX',suffix,compiler)
    IF(odcard(compiler,'OBJ') > 0) THEN
      CALL odget(compiler,'OBJ',cobj) ! the suffix of the main output file
    ELSE
      cobj=obj_suffix
    ENDIF
    
    ! there is at least one target, the object file, even for an include file which is its own target
     
    IF(cobj == ' ') THEN
      CALL odinit(target)
      CALL odinsert(file,'TARGET',target)
      CALL odput(target,'NAME',TRIM(basename))
      CALL odput(target,'FULLNAME',filename) ! the include file is its own target via a "touch" command
    ELSE
      CALL odinit(target)
      CALL odinsert(file,'TARGET',target)
      CALL odput(target,'NAME',TRIM(basename))
      CALL odput(target,'TYPE','OBJECT')
      CALL odfile_name(fullname,bdir,TRIM(basename)//cobj)
      CALL odput(target,'FULLNAME',fullname)
    ENDIF
    
    ! reading #include INCLUDE and USE statements  
      
    IF(suffix(1:2) == '.f' .OR. suffix(1:2) == '.F' .OR. suffix(1:2) == '.c' .OR. suffix(1:2) == '.h') THEN
      
      imod=0 ! module flag needed to catch correctly MODUL PROCEDURE instructions within interfaces
      
      source:DO
      
        ctype=' ' ; ideb=1 ; ifin=0 ; istat=0

        READ(kfile,'(A)',END=10) line
        
        IF(suffix(1:2) == '.f') CALL small(line)
        
        IF(suffix(3:3) /= '9'  .AND. (line(1:1) == '*' .OR. line(1:1) == 'c')) CYCLE

        DO k=1,LEN_TRIM(line)
          car=line(k:k)
          IF(car == ' ' .OR. IACHAR(car) == 8 .OR. car == '#') THEN
            icar=0
          ELSE IF(car == '!') THEN
            icar=1
          ELSE IF(car >= 'a' .AND. car <= 'z') THEN
            icar=2
          ELSE IF(car == '_' .OR. car == '$') THEN
            icar=2
          ELSE IF(car >= '0' .AND. car <= '9') THEN
            icar=3
          ELSE IF(car == "'" .OR. car == '"') THEN !'
            icar=4
          ELSE
            icar=5
          ENDIF
          SELECT CASE(1000*istat+icar)
          CASE(0)
          CASE(1)    ; CYCLE source
          CASE(2)
              IF     (car == 'u') THEN ! possible start of a USE statement
                istat=10
              ELSE IF(car == 'i') THEN ! possible start of a INCLUDE statement
                istat=20
              ELSE IF(car == 'm' .AND. imod == 0) THEN ! possible start of a MODULE statement
                istat=30
              ELSE IF(car == 'e' .AND. imod == 1) THEN ! possible end of a module
                istat=40
              ELSE 
                CYCLE source
              ENDIF
              
          CASE(10002) ; IF(car == 's') THEN ; istat=11 ; ELSE ; CYCLE source ; ENDIF
          CASE(11002) ; IF(car == 'e') THEN ; istat=12 ; ELSE ; CYCLE source ; ENDIF
          CASE(12000) ; istat=13
          CASE(13000)
          CASE(13002) ; istat=14 ; ideb=k ; ifin=k ; ctype='USE'
          CASE(14002) ; istat=14 ; ifin=k
          CASE(14003) ; ifin=k
          CASE(14000,14001,14004,14005) ; EXIT
          
          CASE(20002) ; IF(car == 'n') THEN ; istat=21  ; ELSE ; CYCLE source ; ENDIF
          CASE(21002) ; IF(car == 'c') THEN ; istat=22  ; ELSE ; CYCLE source ; ENDIF
          CASE(22002) ; IF(car == 'l') THEN ; istat=23  ; ELSE ; CYCLE source ; ENDIF
          CASE(23002) ; IF(car == 'u') THEN ; istat=24  ; ELSE ; CYCLE source ; ENDIF
          CASE(24002) ; IF(car == 'd') THEN ; istat=25  ; ELSE ; CYCLE source ; ENDIF
          CASE(25002) ; IF(car == 'e') THEN ; istat=26  ; ELSE ; CYCLE source ; ENDIF
          CASE(26000)
          CASE(26004) ; istat=27
          CASE(27002) ; istat=28; ideb=k
          CASE(28002,28003,28005)
          CASE(28004) ; ifin=k-1 ; ctype='INCLUDE' ; EXIT
          
          CASE(30002) ; IF(car == 'o') THEN ; istat=31  ; ELSE ; CYCLE source ; ENDIF
          CASE(31002) ; IF(car == 'd') THEN ; istat=32  ; ELSE ; CYCLE source ; ENDIF
          CASE(32002) ; IF(car == 'u') THEN ; istat=33  ; ELSE ; CYCLE source ; ENDIF
          CASE(33002) ; IF(car == 'l') THEN ; istat=34  ; ELSE ; CYCLE source ; ENDIF
          CASE(34002) ; IF(car == 'e') THEN ; istat=35  ; ELSE ; CYCLE source ; ENDIF
          CASE(35000) ; istat=36
          CASE(36000)
          CASE(36002) ; istat=37; ideb=k ; ifin=k ; ctype='MODULE'
          CASE(37002,37003) ; ifin=k
          CASE(37000,37001,37004,37005) ; EXIT
          
          CASE(40002) ; IF(car == 'n') THEN ; istat=41  ; ELSE ; CYCLE source ; ENDIF
          CASE(41002) ; IF(car == 'd') THEN ; istat=42  ; ELSE ; CYCLE source ; ENDIF
          CASE(42000) ; istat=42
          CASE(42002) ; IF(car == 'm') THEN ; istat=43  ; ELSE ; CYCLE source ; ENDIF
          CASE(43002) ; IF(car == 'o') THEN ; istat=44  ; ELSE ; CYCLE source ; ENDIF
          CASE(44002) ; IF(car == 'd') THEN ; istat=45  ; ELSE ; CYCLE source ; ENDIF
          CASE(45002) ; IF(car == 'u') THEN ; istat=46  ; ELSE ; CYCLE source ; ENDIF
          CASE(46002) ; IF(car == 'l') THEN ; istat=47  ; ELSE ; CYCLE source ; ENDIF
          CASE(47002) ; IF(car == 'e') THEN ; imod=0           ; CYCLE source ; ENDIF
          
          CASE DEFAULT ; CYCLE source
          
          END SELECT
        ENDDO
        
        IF(ctype == 'MODULE') THEN
          imod=1
          CALL odinit(target)
          CALL odinsert(file,'TARGET',target)
          CALL odinsert(target,'NAME',line(ideb:ifin))
          CALL odinsert(target,'TYPE',ctype)
          CALL odfile_name(fullname,builddir,line(ideb:ifin)//mod_suffix)
          CALL odput(target,'FULLNAME',fullname)
        ELSE IF(ctype /= ' ') THEN
          IF(ctype == 'USE') THEN
            IF(line(ideb:ifin) == 'iso_c_binding')   CYCLE source
            IF(line(ideb:ifin) == 'iso_fortran_env') CYCLE source
            IF(line(ideb:ifin) == 'omp')             CYCLE source
          ENDIF
          CALL odget(file,'DEPEND',line(ideb:ifin),depend)
          IF(.NOT.odisvalid(depend)) THEN
            CALL odinit(depend)
            CALL odinsert(file,'DEPEND',depend)
            CALL odinsert(depend,'NAME',line(ideb:ifin))
            CALL odinsert(depend,'TYPE',ctype)
          ENDIF
        ENDIF
        
      ENDDO source
      
      10 CONTINUE
      
    ENDIF
    
    CLOSE(kfile)
    
  END SUBROUTINE list_target_depend
  
  SUBROUTINE compute_command(filename,suffix,command)
    CHARACTER(*),INTENT(in)  :: filename,suffix
    CHARACTER(*),INTENT(out) :: command
    TYPE(odbase) :: compiler,except
    CHARACTER(255) :: comp,flags
    CALL odget(platform,'COMPILER','SUFFIX',suffix,compiler)
    CALL odget(compiler,'COMMAND',comp)
    CALL odget(compiler,'FLAGS'  ,flags)
    CALL odget(rootcfg,'EXCEPT','FILENAME',filename,except)
    IF(odisvalid(except)) THEN
      CALL odget(except,'FLAGS',flags)
    ENDIF
    command=TRIM(comp)//' '//TRIM(flags)
  END SUBROUTINE
  
  SUBROUTINE small(string)
    CHARACTER(*),INTENT(inout) :: string
    INTEGER   :: i
    CHARACTER :: car
    INTEGER   :: ic,iq
    ic=0;iq=0
    DO i=1,LEN_TRIM(string)
      car=string(i:i)
      IF(car == "'" .AND. iq==0) THEN ! '
        ic=1-ic
      ELSE IF(car == '"' .AND. ic == 0) THEN
        iq=1-iq
      ELSE IF(ic+iq == 0) THEN
        IF(car >= 'A' .AND. car <= 'Z') THEN
          string(i:i)=CHAR(ICHAR(car)+ICHAR('a')-ICHAR('A'))
        ENDIF
      ENDIF
    ENDDO
  END SUBROUTINE small
  
  SUBROUTINE check_external_dependencies(config,task,status)
  
    TYPE(odbase),INTENT(inout) :: config ! current set of tasks
    TYPE(odbase),INTENT(inout) :: task   ! current task
    INTEGER     ,INTENT(out)   :: status
    
    INTEGER                  :: i,j,k,l,nfile,ifound,compiled,ndep
    CHARACTER(255)           :: cname,cfile,cdep,cmod
    CHARACTER(512)           :: fullname
    CHARACTER(8)             :: ctype
    TYPE(odbase)             :: file,subtask,target,depend
    TYPE(odbase),POINTER     :: listfile(:),listdep(:)
    INTEGER     ,POINTER     :: listtask(:)
    
    status=0
    
    CALL odgetptrfamily(task,'FILE'   ,listfile)
    CALL odgetptrfamily(task,'SUBTASK',listtask)
    
    nfile=SIZE(listfile)
    
    DO i=1,nfile
    
      file=listfile(i)
      CALL odget(file,'COMPILED',compiled)
      
      ! The next line is commented because the DEPFILE family must
      ! be computed again : it changes if one inserts a new file in a source directory
      !IF(compiled > 0) CYCLE ! the file has already been analyzed
      
      CALL odgetptrfamily(file,'DEPEND',listdep)
      
      CALL oddelete(file,'DEPFILE')
      
      loop_k : DO k=1,SIZE(listdep)
      
        CALL odget(listdep(k),'TYPE',ctype)
        CALL odget(listdep(k),'NAME',cdep )
        
        IF(ctype == 'INCLUDE') THEN
        
          CALL local_search(task,cdep,fullname,ifound)
          
          IF(ifound == 0) THEN ! file not found locally
          
            CALL external_search(cdep,fullname,ifound)
            
            IF(ifound == 0) THEN
              status=status+1
              CALL odget(file,'NAME',cfile)
              CALL error("impossible to find the include file",cdep,"needed by",cfile)
            ELSE
              CALL odput(listdep(k),'FULLNAME',fullname)
            ENDIF
            
          ELSE
            CALL odput(listdep(k),'FULLNAME',fullname)
            ! looking for the include file index in all the files managed by the task
            DO l=1,nfile
              CALL odget(task,'FILE',depend,l)
              CALL odget(depend,'NAME',cfile)
              IF(cfile == fullname) THEN
                CALL odinsert(file,'DEPFILE',l) ! registering that dependency in the current file
                CALL odput(listdep(k),'DEPFILE',l) ! registering the file index in the dependency
                EXIT
              ENDIF
            ENDDO
          ENDIF
          
        ELSE IF(ctype == 'USE') THEN
        
          ! looking for a task (file with target) generating that module
           
          DO l=1,SIZE(listtask)
            subtask=listfile(listtask(l))
            DO j=1,odcard(subtask,'TARGET')
              CALL odget(subtask,'TARGET',target,j)
              CALL odget(target,'TYPE',ctype)
              IF(ctype == 'MODULE') THEN
                CALL odget(target,'NAME',cmod)
                IF(cmod == cdep) THEN
                  CALL odinsert(file,'DEPFILE',listtask(l)) ! the current file depends on that sub-task
                  CALL odget(target,'FULLNAME',fullname)
                  CALL odput(listdep(k),'FULLNAME',fullname)
                  CYCLE loop_k ! a module depends on a single file
                ENDIF
              ENDIF
            ENDDO
          ENDDO
          
          ! Looking for the module in the build directory
          ! modules outside the projects are ignored
          
          cname=TRIM(cdep)//mod_suffix
        
          CALL build_search(config,cname,fullname,ifound)
          
          IF(ifound /= 0) THEN
            CALL odput(listdep(k),'EXTERNAL',1)
            CALL odput(listdep(k),'FULLNAME',TRIM(fullname))
          ELSE
            CALL odput(listdep(k),'EXTERNAL',2)
            ! starting by a search in local DIRDEP
            CALL local_search(task,cname,fullname,ifound)
            IF(ifound == 0) THEN
              ! continuing by a search in platform DIRDEP
              CALL external_search(cname,fullname,ifound)
              IF(ifound == 0) THEN
                status=status+1
                CALL odget(file,'NAME',cfile)
                CALL error('impossible to find the module file',cdep,'needed by',cfile)
              ELSE
                CALL odput(listdep(k),'FULLNAME',TRIM(fullname))
              ENDIF
            ELSE
              CALL odput(listdep(k),'FULLNAME',TRIM(fullname))
            ENDIF
          ENDIF

       ENDIF  
          
      ENDDO loop_k
      
    ENDDO
    
  END SUBROUTINE check_external_dependencies
  
  SUBROUTINE local_search(task,cdep,fullname,ifound)
  
    TYPE(odbase),INTENT(in)  :: task
    CHARACTER(*),INTENT(in)  :: cdep
    CHARACTER(*),INTENT(out) :: fullname
    INTEGER     ,INTENT(out) :: ifound
    
    INTEGER        :: k
    CHARACTER(512) :: dir
    
    ifound=0
    
    CALL odfile_name(fullname,builddir,cdep)
    
    IF(odfile_exist(fullname)) THEN
      ifound=1
      RETURN
    ENDIF
    
    DO k=1,odcard(task,'DIR')
      CALL odget(task,'DIR',dir,k)
      CALL odfile_name(fullname,dir,cdep)
      IF(odfile_exist(fullname)) THEN
        ifound=1
        RETURN
      ENDIF
    ENDDO
    
    DO k=1,odcard(task,'DIRDEP')
      CALL odget(task,'DIRDEP',dir,k)
      CALL odfile_name(fullname,dir,cdep)
      IF(odfile_exist(fullname)) THEN
        ifound=1
        RETURN
      ENDIF
    ENDDO
    
  END SUBROUTINE local_search
  
  SUBROUTINE build_search(config,cdep,fullname,ifound)
  
    TYPE(odbase),INTENT(in)  :: config   ! set of tasks
    CHARACTER(*),INTENT(in)  :: cdep     ! basename + suffix
    CHARACTER(*),INTENT(out) :: fullname ! full name relative to the root of the project
    INTEGER     ,INTENT(out) :: ifound   ! 0=not found, 1=local to the build of task, 2=another task target
    
    INTEGER                :: k
    CHARACTER(512)         :: dir,cdir
    TYPE(odbase)  ,POINTER :: listtask(:)
    
    ifound=0
    
    CALL odfile_name(fullname,builddir,cdep)
      
    IF(odfile_exist(fullname)) THEN
      ifound=1
      RETURN
    ENDIF
    
    CALL odgetptrfamily(config,'TASK',listtask)
    
    DO k=1,SIZE(listtask)

      IF(odcard(listtask(k),'BUILD') == 0) CYCLE
      
      CALL odget(listtask(k),'BUILD',dir)
      CALL odfile_name(cdir,builddir,dir)
      CALL odfile_name(fullname,cdir,cdep)
      
      IF(odfile_exist(fullname)) THEN
        ifound=1
        RETURN
      ENDIF

    ENDDO
    
  END SUBROUTINE build_search
  
  SUBROUTINE external_search(cdep,fullname,ifound)
  
    CHARACTER(*),INTENT(in)  :: cdep
    CHARACTER(*),INTENT(out) :: fullname
    INTEGER     ,INTENT(out) :: ifound
    
    INTEGER        :: k
    CHARACTER(512) :: dir
    
    ifound=0
    
    DO k=1,odcard(platform,'DIRDEP') ! directories for external includes or modules
      CALL odget(platform,'DIRDEP',dir,k)
      CALL odfile_name(fullname,dir,cdep)
      IF(odfile_exist(fullname)) THEN
        ifound=1
        RETURN
      ENDIF
    ENDDO
    
  END SUBROUTINE external_search
  
  SUBROUTINE compute_file_level(task,level,status)
  
    TYPE(odbase),INTENT(in)    :: task
    INTEGER     ,INTENT(out)   :: level(:)
    INTEGER     ,INTENT(inout) :: status
    
    TYPE(odbase)         :: file
    INTEGER              :: i,j,k,n,lev,mod
    TYPE(odbase),POINTER :: listfile(:)
    INTEGER     ,POINTER :: listdep(:)
    
    CALL odgetptrfamily(task,'FILE'   ,listfile)
    n=SIZE(listfile)
    IF(n==0) RETURN
    
    level=1 ! all the sub-tasks have the level 1 by default
    
    DO i=1,n
    
      mod = 0
      
      DO j=1,n
        
        file=listfile(j)
        CALL odgetptrfamily(file,'DEPFILE',listdep)
        lev=0
        DO k=1,SIZE(listdep)
          IF(listdep(k) > 0) THEN
            lev=MAX(lev,level(listdep(k)))
          ENDIF
        ENDDO
        IF(level(j) <= lev) THEN
          mod=mod+1
          level(j)=lev+1
        ENDIF
        
      ENDDO
      
      IF(mod == 0) EXIT
      
    ENDDO
    
    IF(i > n+1) THEN
      CALL error('recursivity between dependencies')
      status=status+1
    ENDIF
    
  END SUBROUTINE compute_file_level
  
  SUBROUTINE compile_file(task,file,status,niv)
  
    TYPE(odbase),INTENT(in)    :: task
    TYPE(odbase),INTENT(inout) :: file
    INTEGER     ,INTENT(out)   :: status
    INTEGER     ,INTENT(in)    :: niv
    
    TYPE(odbase)   :: depend,target,compiler,except
    CHARACTER(255) :: command
    CHARACTER(5000):: arguments
    CHARACTER(512) :: fullname,bdir,listinc(1000),dirname
    CHARACTER(255) :: basename,text,flags,object,cfile,cname
    CHARACTER(8)   :: suffix,cinc,cmod,cout,cnl,csuff,mode,ctype
    INTEGER        :: sig
    CHARACTER      :: c
    INTEGER        :: k,l,time,timedep,n,ninc
    
    status=0
    
    CALL odget(task,'BUILD',text)
    CALL odfile_name(bdir,builddir,text)
    
    CALL odget(file,'BASENAME',basename)
    CALL odget(file,'NAME'    ,cfile)
    CALL odget(file,'SUFFIX'  ,suffix)
    
    ! specific treatment of include files
    
    loop_k : DO k=1,odcard(platform,'INC')
      CALL odget(platform,'INC',csuff,k)
      IF(suffix == csuff) THEN
        ! it is advise to check time stamps
        CALL odfile_info(cfile,time=time)
        DO l=1,odcard(file,'DEPEND')
          CALL odget(file,'DEPEND',depend,l)
          CALL odget(depend,'FULLNAME',fullname)
          CALL odfile_info(fullname,time=timedep)
          CALL odput(depend,'TIME',time)
          IF(timedep > time) EXIT loop_k
        ENDDO
        CALL odput(file,'COMPILED',1)
        CALL odput(file,'TIME',time)
        RETURN
      ENDIF
    ENDDO loop_k
    
    ! normal compilation
    
    CALL odget(platform,'COMPILER','SUFFIX',suffix,compiler) ! the compiler options
    CALL odget(compiler,'MODE'   ,mode)    ! the compiler options
    CALL odget(compiler,'COMMAND',command) ! compiler command 
    CALL odget(compiler,'FLAGS' ,flags)    ! compiler flags
    
    CALL odget(rootcfg,'EXCEPT','FILENAME',cfile,except)
    
    IF(odisvalid(except)) THEN
      CALL odget(except,'FLAGS',flags)
    ENDIF
    
    CALL odput(file,'COMMAND',TRIM(command)//' '//TRIM(flags)) ! to remember the command used
    
    IF(mode == 'FILTER') THEN
    
      CALL odget(file,'TARGET',target)
      CALL odget(target,'FULLNAME',fullname)
      arguments=TRIM(flags)//' '//TRIM(cfile)//' '//TRIM(fullname)

    ELSE

      CALL odget(compiler,'INC'    ,cinc)  ! the include/module directory flag
      CALL odget(compiler,'MOD'    ,cmod)  ! the output module directory flag
      CALL odget(compiler,'OUT'    ,cout)  ! the ouput flag
      CALL odget(compiler,'NOLINK' ,cnl)   ! the "compile only" flag

      arguments=TRIM(flags)//' '//cnl
      
      ninc=0
      
      IF(cinc /= ' ') THEN
      
        n=LEN_TRIM(cinc) ; c=cinc(n:n)
        IF(c == '_') THEN
          n=n-1
        ELSE IF(c /= ':' .AND. c /= '=') THEN
          n=n+1
        ENDIF
        
        DO k=1,odcard(platform,'DIRDEP')
          CALL odget(platform,'DIRDEP',listinc(k),k)
          arguments=TRIM(arguments)//' '//cinc(1:n)//listinc(k)
        ENDDO
        
        ninc=k

        DO k=1,odcard(file,'DEPEND')
          CALL odget(file,'DEPEND',depend,k)
          CALL odget(depend,'NAME',cname)
          CALL odget(depend,'TYPE',ctype)
          CALL odget(depend,'FULLNAME',fullname)
          IF(ctype == 'USE') THEN
            dirname=fullname(1:LEN_TRIM(fullname)-LEN_TRIM(cname)-LEN_TRIM(mod_suffix)-1)
          ELSE
            dirname=fullname(1:LEN_TRIM(fullname)-LEN_TRIM(cname)-1)
          ENDIF
          IF(dirname /= ' ') THEN
            DO l=1,ninc
              IF(dirname == listinc(l)) GOTO 10
            ENDDO
              ninc=ninc+1
              IF(ninc < SIZE(listinc)) THEN
                ninc=ninc+1
                listinc(ninc)=dirname
              ENDIF
              arguments=TRIM(arguments)//' '//cinc(1:n)//TRIM(dirname)
  10        CONTINUE
          ENDIF
        ENDDO
        
        DO k=1,odcard(task,'DIRDEP')
          CALL odget(task,'DIRDEP',dirname,k)
          IF(dirname /= ' ') THEN
            DO l=1,ninc
              IF(dirname == listinc(l)) GOTO 20
            ENDDO
              ninc=ninc+1
              IF(ninc < SIZE(listinc)) THEN
                ninc=ninc+1
                listinc(ninc)=dirname
              ENDIF
              arguments=TRIM(arguments)//' '//cinc(1:n)//TRIM(dirname)
  20        CONTINUE
          ENDIF
        ENDDO
        
      ENDIF

      ! location of other targets : the root build directory

      IF(cmod /= ' ') THEN
        n=LEN_TRIM(cmod) ; c=cmod(n:n)
        IF(c == '_') THEN
          n=n-1
        ELSE IF(c /= ':' .AND. c /= '=') THEN
          n=n+1
        ENDIF
        arguments=TRIM(arguments)//' '//cmod(1:n)//TRIM(builddir)
      ENDIF

      ! source file

      arguments=TRIM(arguments)//' '//TRIM(cfile)

      ! object target

      IF(cout /= ' ') THEN
        n=LEN_TRIM(cout) ; c=cout(n:n)
        IF(c == '_') THEN
          n=n-1
        ELSE IF(c /= ':' .AND. c /= '=') THEN
          n=n+1
        ENDIF
        CALL odfile_name(object,bdir,TRIM(basename)//obj_suffix)
        arguments=TRIM(arguments)//' '//cout(1:n)//TRIM(object)
      ENDIF
      
    ENDIF
    
    ! launching the command
    
    CALL message(niv,command,arguments)
    CALL odsystem(command,arguments,status)
    
    ! checking the creation of targets
    
    DO k=1,odcard(file,'TARGET')
      CALL odget(file,'TARGET',target,k)
      CALL odget(target,'FULLNAME',fullname)
      IF(.NOT.odfile_exist(fullname)) THEN
        status=status+1
      ELSE
        CALL odfile_info(fullname,time=time,sig=sig)
        CALL odput(target,'TIME',time)
        CALL odput(target,'SIG',sig)
      ENDIF
    ENDDO
    
    IF(status == 0) THEN 
    
       ! all targets have been generated. One has to register dependency information
       
      DO k=1,odcard(file,'DEPEND')
        CALL odget(file,'DEPEND',depend,k)
        CALL odget(depend,'FULLNAME',fullname)
        CALL odfile_info(fullname,time=time,sig=sig)
        CALL odput(depend,'TIME',time)
        CALL odput(depend,'SIG',sig)
      ENDDO
      
      ! idem for the main file
      
      CALL odget(file,'NAME',fullname)
      CALL odfile_info(fullname,time=time,sig=sig)
      CALL odput(file,'TIME',time)
      CALL odput(file,'SIG',sig)
      CALL odput(file,'COMPILED',1)
      
    ELSE
    
      CALL odput(file,'COMPILED',0) ! compilation failure
      CALL odget(file,'NAME',fullname)
      CALL error('compilation failure of the file',fullname)
      
    ENDIF
  
  END SUBROUTINE compile_file
  
  SUBROUTINE message(niv,m1,m2,m3,m4,m5)
    ! print out inside a critical region because WRITE is not thread safe
    INTEGER     ,INTENT(in)          :: niv ! task level
    CHARACTER(*),INTENT(in)          :: m1
    CHARACTER(*),INTENT(in),OPTIONAL :: m2,m3,m4,m5
    CHARACTER(100) :: s=' '
    !$OMP CRITICAL
    IF(PRESENT(m5)) THEN
      WRITE(*,"(6(1X,A))") s(1:2*niv-1),TRIM(m1),TRIM(m2),TRIM(m3),TRIM(m4),TRIM(m5)
    ELSE IF(PRESENT(m4)) THEN
      WRITE(*,"(6(1X,A))") s(1:2*niv-1),TRIM(m1),TRIM(m2),TRIM(m3),TRIM(m4)
    ELSE IF(PRESENT(m3)) THEN
      WRITE(*,"(6(1X,A))") s(1:2*niv-1),TRIM(m1),TRIM(m2),TRIM(m3)
    ELSE IF(PRESENT(m2)) THEN
      WRITE(*,"(6(1X,A))") s(1:2*niv-1),TRIM(m1),TRIM(m2)
    ELSE
      WRITE(*,"(6(1X,A))") s(1:2*niv-1),TRIM(m1)
    ENDIF
    FLUSH(6)
    !$OMP END CRITICAL
  END SUBROUTINE
  
  SUBROUTINE error(m1,m2,m3,m4,m5)
    ! print out inside a critical region because WRITE is not thread safe
    ! and because "nerror" must be protected
    CHARACTER(*),INTENT(in)          :: m1
    CHARACTER(*),INTENT(in),OPTIONAL :: m2,m3,m4,m5
    CHARACTER(150),ALLOCATABLE :: cmess(:)
    !$OMP CRITICAL
    IF(nerror == SIZE(errmes)) THEN
      ALLOCATE(cmess(2*nerror))
      cmess(1:SIZE(errmes))=errmes
      DEALLOCATE(errmes)
      CALL move_alloc(cmess,errmes)
    ENDIF
    nerror=nerror+1
    IF(PRESENT(m5)) THEN
      WRITE(errmes(nerror),'(5(1X,A))') TRIM(m1),TRIM(m2),TRIM(m3),TRIM(m4),TRIM(m5)
    ELSE IF(PRESENT(m4)) THEN
      WRITE(errmes(nerror),'(5(1X,A))') TRIM(m1),TRIM(m2),TRIM(m3),TRIM(m4)
    ELSE IF(PRESENT(m3)) THEN
      WRITE(errmes(nerror),'(5(1X,A))') TRIM(m1),TRIM(m2),TRIM(m3)
    ELSE IF(PRESENT(m2)) THEN
      WRITE(errmes(nerror),'(5(1X,A))') TRIM(m1),TRIM(m2)
    ELSE
      WRITE(errmes(nerror),'(5(1X,A))') TRIM(m1)
    ENDIF
    WRITE(*,'(A)') TRIM(errmes(nerror))
    FLUSH(6)
    !$OMP END CRITICAL
  END SUBROUTINE
  
  SUBROUTINE print_manual
  
    WRITE(*,'(1X,A)') " "
    WRITE(*,'(1X,A)') "Windows : "
    WRITE(*,'(1X,A)') " "
    WRITE(*,'(1X,A)') "  %builder% [-h] [-v] [-d project_directory] [-f configuration_file] [platform_name]"
    WRITE(*,'(1X,A)') ""
    WRITE(*,'(1X,A)') "Linux : "
    WRITE(*,'(1X,A)') " "
    WRITE(*,'(1X,A)') "   builder  [-h] [-v] [-d project_directory] [-f configuration_file] [platform_name]"
    WRITE(*,'(1X,A)') ""
    WRITE(*,'(1X,A)') " -h : short user manual"
    WRITE(*,'(1X,A)') " -v : verbose mode, the builder explains its work"
    WRITE(*,'(1X,A)') " -d : providing the project root directory. Default : current directory."
    WRITE(*,'(1X,A)') " -f : providing the configuration file. default ; builder.cfg"
    WRITE(*,'(1X,A)') " "
    WRITE(*,'(1X,A)') "If the platform name is missing, then the first platform of the configuration file is chosen."
    WRITE(*,'(1X,A)') " " 
    WRITE(*,'(1X,A)') "Configuration file : "
    WRITE(*,'(1X,A)') " " 
    WRITE(*,'(1X,A)') "   *-> EXCEPT         s     ! exception"
    WRITE(*,'(1X,A)') "       -> FLAGS       c0|t" ! compilation flag"
    WRITE(*,'(1X,A)') "       -> FILENAME    c0|t" ! source file name"
    WRITE(*,'(1X,A)') "       -> PLATFORM    c0|t" ! particular platform
    WRITE(*,'(1X,A)') " " 
    WRITE(*,'(1X,A)') "   *-> PLATFORM       s"    ! general options associated to an operating system and a set of compilers"
    WRITE(*,'(1X,A)') "       -> BUILD       c0|t  ! directory receiving constructed files"
    WRITE(*,'(1X,A)') "      *-> COMPILER          ! compiler description"
    WRITE(*,'(1X,A)') "          -> COMMAND  c0|t  ! compiler command  "   
    WRITE(*,'(1X,A)') "          -> FLAGS    c0|t  ! compiler flag"
    WRITE(*,'(1X,A)') "          -> OUT      c0|t  ! flag introducing the bject file name"
    WRITE(*,'(1X,A)') "          -> INC      c0|t  ! flag followed by an include directory"
    WRITE(*,'(1X,A)') "          -> MOD      c0|t  ! flags followed by the directory receiving modules"
    WRITE(*,'(1X,A)') "          -> SUFFIX   C0    ! suffixes of the source files compiled with that compiler"
    WRITE(*,'(1X,A)') "      *-> DIRDEP      c0|t  ! external dependency directory"
    WRITE(*,'(1X,A)') "       -> DYN               ! dynamic library suffix"
    WRITE(*,'(1X,A)') "       -> DYNLIB            ! dynamic library"
    WRITE(*,'(1X,A)') "          -> COMMAND  c0|t  ! main command"
    WRITE(*,'(1X,A)') "          -> FLAGS    c0|t  ! possible flags"
    WRITE(*,'(1X,A)') "          -> OUT      c0|t  ! flag introducing the library name"
    WRITE(*,'(1X,A)') "         [-> MODE     i0    ! =2 to repeat the library name as first object file"
    WRITE(*,'(1X,A)') "       -> DYNPRE      c0    ! dynamic library prefix"
    WRITE(*,'(1X,A)') "       -> EXE         c0    ! executable suffix"
    WRITE(*,'(1X,A)') "       -> LIB         c0    ! library suffix"
    WRITE(*,'(1X,A)') "       -> LIBPRE      c0    ! library prefix"
    WRITE(*,'(1X,A)') "       -> LIBRARY           ! archive description"
    WRITE(*,'(1X,A)') "          -> COMMAND  c0|t  ! main command"
    WRITE(*,'(1X,A)') "          -> FLAGS    c0|t  ! possible flags"
    WRITE(*,'(1X,A)') "          -> OUT      c0|t  ! flag introducing the library name"
    WRITE(*,'(1X,A)') "         [-> MODE     i0    ! =2 to repeat the library name as first object file"
    WRITE(*,'(1X,A)') "       -> NAME        c0|t  ! platform name"
    WRITE(*,'(1X,A)') "       -> MOD         c0    ! module suffix"
    WRITE(*,'(1X,A)') "       -> OBJ         c0    ! object suffix"
    WRITE(*,'(1X,A)') "       -> PROGRAM           ! linker"
    WRITE(*,'(1X,A)') "          -> COMMAND  c0|t  ! main command"
    WRITE(*,'(1X,A)') "          -> FLAGS    c0|t  ! various flags"
    WRITE(*,'(1X,A)') "          -> OUT      c0|t  ! the flags introducing the program file name"
    WRITE(*,'(1X,A)') " " 
    WRITE(*,'(1X,A)') "   *-> TASK           s     ! part of the work"
    WRITE(*,'(1X,A)') "     [ -> BUILD       t     ! path relative to the root build directory to save object files"
    WRITE(*,'(1X,A)') "     [*-> COND        s     ! conditional block"
    WRITE(*,'(1X,A)') "         *-> IF       i0    ! condition"
    WRITE(*,'(1X,A)') "          -> ...            ! any information of the TASK block"
    WRITE(*,'(1X,A)') "     [*-> DEPEND      c0|t  ! possible TASK dependency (name of another task)"
    WRITE(*,'(1X,A)') "     [*-> DEPFILE     c0|t  ! external file dependency (relative or absolute file name)"
    WRITE(*,'(1X,A)') "     [*-> DIR         t     ! directory path to locate source files"
    WRITE(*,'(1X,A)') "     [*-> DIRDEP      c0|t  ! directory where looking for dependencies"
    WRITE(*,'(1X,A)') "     [*-> DIROBJ      c0|t  ! directory where getting additional object files"
    WRITE(*,'(1X,A)') "     [*-> LIBRARY     c0|t  ! name of a library needed to make the program"
    WRITE(*,'(1X,A)') "     [ -> LIBNAME     c0|t  ! name of the library to create"
    WRITE(*,'(1X,A)') "     [*-> INST        t     ! instructions to execute"
    WRITE(*,'(1X,A)') "     [*-> FILENAME    t     ! specific files to compile"
    WRITE(*,'(1X,A)') "     [*-> DYNLIB      c0|t  ! name of a dynamic library needed to make the program"
    WRITE(*,'(1X,A)') "     [ -> DYNNAME     c0|t  ! name of the dynamic library to create"
    WRITE(*,'(1X,A)') "       -> NAME        c0|t  ! the name of a dependency task"
    WRITE(*,'(1X,A)') "     [ -> PROGRAM     c0|t  ! name of the executable program"
    WRITE(*,'(1X,A)') "     [*-> SUFFIX      c0|t  ! explicit source file suffixes, in conjunction with source directories"
    WRITE(*,'(1X,A)') "     [*-> TARGET      t     ! target file associated to instructions"
    WRITE(*,'(1X,A)') "     [*-> TASK        s     ! sub task executed before compiling the task source files"
    WRITE(*,'(1X,A)') "          -> ..."
    WRITE(*,'(1X,A)') " "
    WRITE(*,'(1X,A)') "Notations :"
    WRITE(*,'(1X,A)') "  c0 : short string (up to 8 characters), possibly delimited by apostrophes"
    WRITE(*,'(1X,A)') "  t  : long string between quotation marks"
    WRITE(*,'(1X,A)') "  i0 : scalar integer"
    WRITE(*,'(1X,A)') "  s  : data block starting by STRUCTURE and finishing by END"
    WRITE(*,'(1X,A)') "  [  : optional data"
    WRITE(*,'(1X,A)') "  *  : several instances are authorized"
    FLUSH(6)

  END SUBROUTINE
    
END MODULE
